package com.att.gs.codip.ograph.examples;

import java.io.Reader;

import jess.Jesp;
import jess.JessException;
import jess.Rete;
import jess.Userpackage;

/** **********************************************************************
 * A command-line interface for Jess
 * <P>
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 * <P>
 * This is a lightly modified verison of the jess.Main class. It can be 
 * imbedded into an application to provide a jess command line interface
 * primarily for debuging.
 ********************************************************************** */

public class JessMainLoop {

    private Rete m_rete;
    private Reader m_fis;
    private Jesp m_j;
    private boolean m_readStdin = true;

    /**
     * Getter for Rete engine.
     **/
    public Rete getRete() { return m_rete; }
    
    
    
    /**
     * Display the Jess startup banner on the Rete object's standard
     * output, something like
     *
     * <PRE>
     * Copyright (C) 1998 E.J. Friedman Hill and the Sandia Corporation
     * Jess Version 5.0a5 5/27/99
     * </PRE>
     * */

    public void showLogo()
    {
        if (m_rete != null && m_rete.getOutStream() != null) {
            m_rete.getOutStream().println("\nJess, the Java Expert System Shell");
            m_rete.getOutStream().println("Copyright (C) 1998 E.J. Friedman Hill"
                                          + " and the Sandia Corporation");
            try {
                m_rete.executeCommand("(printout t (jess-version-string) crlf crlf)");
            } catch(JessException re) {}
        }
    }
    
    /**
     * Set a Main object up for later execution.
     * @param argv Command-line arguments
     * @param r An initialized Rete object, with routers set up
     */
    public JessMainLoop initialize(Rete r) {
        m_rete = r;        
        m_fis = m_rete.getInputRouter("t");

        loadOptionalFunctions();
        return this ;           
    }
    
    private void loadOptionalFunctions() {
        // **********************************************************************
        // Load in optional packages, but don't fail if any are missing.
        String [] packages = { "jess.BagFunctions",
                               "jess.ViewFunctions" };
        
        for (int i=0; i< packages.length; i++) {
            try {
                m_rete.addUserpackage((Userpackage)
                                      Class.forName(packages[i]).newInstance());
            }
            catch (Throwable t) { /* Optional package not present, OK */ }
        }
        
    }
    
    /**
     * Repeatedly parse and excute commands, from location determined
     * during initialize().
     * @param doPrompt True if a prompt should be printed, false otherwise.
     * Prompts will never be printed during a (batch) command. 
     */
    
    public void execute(boolean doPrompt) {
        // **********************************************************************
        // Process input from file or keyboard
        showLogo();
       
        if (m_fis != null) {
            m_j = new Jesp(m_fis, m_rete);
            do {
                try {
                    // Argument is 'true' for prompting, false otherwise
                    m_j.parse(doPrompt && m_readStdin);

                } catch (JessException re) {
                    if (re.getNextException() != null) {
                        m_rete.getErrStream().write(re.toString());                    
                        m_rete.getErrStream().write("\nNested exception is:\n");    
                        m_rete.getErrStream().println(re.getNextException().getMessage());
                        re.getNextException().printStackTrace(m_rete.getErrStream());

                    } else
                        re.printStackTrace(m_rete.getErrStream());
                } catch (Exception e) {
                    m_rete.getErrStream().println("Unexpected exception:");
                    e.printStackTrace(m_rete.getErrStream());
                } finally {
                    m_rete.getErrStream().flush();
                    m_rete.getOutStream().flush();                
                }
            }
            // Loop if we're using the command line
            while (m_readStdin);
        }
        // If called again, read stdin, not batch file    
        m_readStdin = true;
        m_fis = m_rete.getInputRouter("t");
    }
    
}




