/*
 * @(#)OGOntology.java
 *
 * Ontology Graph API,  Copyright 2000, 2001 GRCI All Rights Reserved.
 *
 */
package com.att.gs.codip.ograph;

import java.util.Collection;
import java.util.Iterator;

/** **************************
 * This is the interface for the ontology container class.
 ************************** **/
public interface OGOntology extends OGResource {


/************************************************
 * Simple Access Methods.
 ************************************************/

/**
 * Use this method to convert an Ontology to a new URI
 */
public void morph(String URI);

/**
 * Setter method for the source URI of this ontology.
 * Public wrapper method for the portected method OGResource.setURI(String) method
 * which is used internally.
 *
 * @param theSourceURI The source URI String.
 **/
public void setSourceURI(String theSourceURI);
/**
 * Getter method for the source URI of this ontology.
 * Provided for completeness returns the same URI string as
 * getURI()
 *
 * @return  The source URI String.
 **/
public String getSourceURI();

/**
 * Setter method for the version information of this ontology.
 *
 * @param theVersionInfo The version information String.
 **/
public void setVersionInfo(String theVersionInfo);
/**
 * Getter method for the version information  of this ontology.
 *
 * @return The version information  String.
 **/
public String getVersionInfo ();

/********************************************
 * Collection Access Methods.
 *******************************************/

/*
 * ---------- Class Methods --------------
 */

/**
 * Looks for a matching class in this ontology.
 * Searches this onlology and all imported ontologies for a class
 * whose ID matches the provided ID.
 *
 * @param theID The ID of the sought class.
 * @return The matching class if found, otherwise returns void.
 **/
public OGClass lookupClass(Object theID);

/**
 * Looks for, and possibilly creates, a Class in this ontology.
 * Searches this onlology and all imported ontologies for a class
 * whose ID matches the provided ID. May create a new instance with
 * the appropriate ID in this ontology if none exists.
 *
 * @param theID The ID of the sought class.
 * @param pCreate true causes a new instance to be created if no existing
 * match is found.
 *
 * @return The matching class if found or created, otherwise returns void.
 **/
public OGClass lookupClass(Object theID, boolean pCreate);

/**
 * Creates a Class in this ontology with a generated unique ID
 * and adds the Class to the ontology.
 *
 * @return The matching class created, otherwise returns void.
 **/
public OGClass createClass();
/**
 * Looks for, and possibilly creates, a Class in this ontology.
 * Searches this onlology and all imported ontologies for a class
 * whose ID matches the provided ID. May create a new instance with
 * the appropriate ID in this ontology if none exists.
 * The class if new is added to the Ontology.
 *
 * @param theID The ID of the new class.
 *
 * @return The matching class if found or created, otherwise returns void.
 **/
public OGClass createClass(String theID);
/**
 * Creates a Class in this ontology with a generated unique ID,
 * sets the Clases type to the passed in type
 * and adds the Class to the ontology.
 *
 * @param aType The type of the new class.
 *
 * @return The matching class created, otherwise returns void.
 **/
public OGClass createClass(OGResource aType);
/**
 * Looks for, and possibilly creates, a Class in this ontology.
 * Searches this onlology and all imported ontologies for a class
 * whose ID matches the provided ID. May create a new instance with
 * the appropriate ID in this ontology if none exists.
 * The class if new is added to the Ontology with the type
 * sets to the passed in type.
 *
 * @param aType The type of the new class.
 * @param theID The ID of the new class.
 *
 * @return The matching class if found or created, otherwise returns void.
 **/
public OGClass createClass(OGResource aType,String theID);

/**
 * Adds the provided class to the ontology.
 *
 * The provided class should be added so that it will be included in
 * both  class and resource collections as appropriate.
 *
 * @param theClass The class to add to the ontology.
 **/
public void addClass(OGClass theClass);

/**
 * Removes the provided class from the ontology.
 *
 * @param theClass The class to remove to the ontology.
 **/
public void removeClass(OGClass theClass);

/**
 * Looks for a Class in this ontology whose ID matches the provided ID.
 *
 * Differs from lookupClass in that the imports are not searched.
 *
 * @param theID The ID of the sought class.
 * @return The matching class if found, otherwise returns void.
 **/
public OGClass getClass(Object theID);

/**
 * Checks for the existence of a Class in this ontology
 * whose ID matches the provided ID.
 *
 * The class must be in this ontology, not in its imports.
 *
 * @param theID The ID of the sought class.
 * @return true if the class exists in this ontology.
 **/
public boolean containsClass(Object theID);

/**
 * Provides a iterator for the all of the ontology's classes.
 *
 * @return An iterator over all the ontology's classes.
 **/
public Iterator classes();

/**
 * Provides a iterator for only the ontology's classes
 * of the provided type.
 *
 * @param aType The type of classes to include in the iterator.
 * @return An iterator over some of the ontology's classes.
 **/
public Iterator classes(OGResource aType);

/**
 * Provides a Collection of the ontology's classes
 * of the provided type.
 *
 * @param aType The type of classes to include in the iterator.
 * @return A collection of some of the ontology's classes.
 **/
public Collection getClasses(OGResource aType);

/*
 * ---------- Datatype Methods --------------
 */

/**
 * Looks for a matching data type in this ontology.
 * Searches this onlology and all imported ontologies for a data type
 * whose ID matches the provided ID.
 *
 * @param theID The ID of the sought data type.
 * @return The matching data type if found, otherwise returns void.
 **/
public OGDatatype lookupDatatype(Object theID);

/**
 * Looks for, and possibilly creates, a Datatype in this ontology.
 * Searches this onlology and all imported ontologies for a data type
 * whose ID matches the provided ID. May create a new instance with
 * the appropriate ID in this ontology if none exists.
 *
 * @param theID The ID of the sought data type.
 * @param pCreate true causes a new instance to be created if no existing
 * match is found.
 *
 * @return The matching data type if found or created, otherwise returns void.
 **/
public OGDatatype lookupDatatype(Object theID, boolean pCreate);

/**
 * Creates a Datatype in this ontology with a generated unique ID
 * and adds the Datatype to the ontology.
 *
 * @return The matching Datatype created, otherwise returns void.
 **/
public OGDatatype createDatatype();
/**
 * Looks for, and possibilly creates, a Datatype in this ontology.
 * Searches this onlology and all imported ontologies for a Datatype
 * whose ID matches the provided ID. May create a new instance with
 * the appropriate ID in this ontology if none exists.
 * The Datatype if new is added to the Ontology.
 *
 * @param theID The ID of the new Datatype.
 *
 * @return The matching Datatype if found or created, otherwise returns void.
 **/
public OGDatatype createDatatype(String theID);
/**
 * Creates a Datatype in this ontology with a generated unique ID,
 * sets the Clases type to the passed in type
 * and adds the Datatype to the ontology.
 *
 * @param aType The type of the new Datatype.
 *
 * @return The matching Datatype created, otherwise returns void.
 **/
public OGDatatype createDatatype(OGResource aType);
/**
 * Looks for, and possibilly creates, a Datatype in this ontology.
 * Searches this onlology and all imported ontologies for a Datatype
 * whose ID matches the provided ID. May create a new instance with
 * the appropriate ID in this ontology if none exists.
 * The Datatype if new is added to the Ontology with the type
 * sets to the passed in type.
 *
 * @param aType The type of the new Datatype.
 * @param theID The ID of the new Datatype.
 *
 * @return The matching Datatype if found or created, otherwise returns void.
 **/
public OGDatatype createDatatype(OGResource aType,String theID);

/**
 * Adds the provided data type to the ontology.
 *
 * The provided data type should be added so that it will be included in
 * both  data type and resource collections as appropriate.
 *
 * @param theDatatype The data type to add to the ontology.
 **/
public void addDatatype(OGDatatype theDatatype);

/**
 * Removes the provided data type from the ontology.
 *
 * @param theDatatype The data type to remove to the ontology.
 **/
public void removeDatatype(OGDatatype theDatatype);

/**
 * Looks for a Datatype in this ontology whose ID matches the provided ID.
 *
 * Differs from lookupDatatype in that the imports are not searched.
 *
 * @param theID The ID of the sought data type.
 * @return The matching data type if found, otherwise returns void.
 **/
public OGDatatype getDatatype(Object theID);

/**
 * Checks for the existence of a Datatype in this ontology
 * whose ID matches the provided ID.
 *
 * The data type must be in this ontology, not in its imports.
 *
 * @param theID The ID of the sought data type.
 * @return true if the data type exists in this ontology.
 **/
public boolean containsDatatype(Object theID);

/**
 * Provides a iterator for the all of the ontology's data types.
 *
 * @return An iterator over all the ontology's data types.
 **/
public Iterator datatypes();

/**
 * Provides a iterator for only the ontology's data types
 * of the provided type.
 *
 * @param aType The type of data types to include in the iterator.
 * @return An iterator over some of the ontology's data types.
 **/
public Iterator datatypes(OGResource aType);

/**
 * Provides a Collection of the ontology's data types
 * of the provided type.
 *
 * @param aType The type of data types to include in the iterator.
 * @return A collection of some of the ontology's data types.
 **/
public Collection getDatatypes(OGResource aType);

/*
 * ---------- Property Methods --------------
 */
/**
 * Looks for a matching property in this ontology.
 * Searches this onlology and all imported ontologies for a property
 * whose ID matches the provided ID.
 *
 * @param theID The ID of the sought property.
 * @return The matching property if found, otherwise returns void.
 **/
public OGProperty lookupProperty(Object theID);

/**
 * Looks for, and possibilly creates, a property in this ontology.
 * Searches this onlology and all imported ontologies for a property
 * whose ID matches the provided ID. May create a new instance with
 * the appropriate ID in this ontology if none exists.
 *
 * @param theID The ID of the sought property.
 * @param pCreate true causes a new instance to be created if no existing
 * match is found.
 *
 * @return The matching property if found or created, otherwise returns void.
 **/
public OGProperty lookupProperty(Object theID, boolean pCreate);

/**
 * Creates a Property in this ontology with a generated unique ID
 * and adds the Property to the ontology.
 *
 * @return The matching Property created, otherwise returns void.
 **/
public OGProperty createProperty();
/**
 * Looks for, and possibilly creates, a Property in this ontology.
 * Searches this onlology and all imported ontologies for a Property
 * whose ID matches the provided ID. May create a new instance with
 * the appropriate ID in this ontology if none exists.
 * The Property if new is added to the Ontology.
 *
 * @param theID The ID of the new Property.
 *
 * @return The matching Property if found or created, otherwise returns void.
 **/
public OGProperty createProperty(String theID);
/**
 * Creates a Property in this ontology with a generated unique ID,
 * sets the Clases type to the passed in type
 * and adds the Property to the ontology.
 *
 * @param aType The type of the new Property.
 *
 * @return The matching Property created, otherwise returns void.
 **/
public OGProperty createProperty(OGResource aType);
/**
 * Looks for, and possibilly creates, a Property in this ontology.
 * Searches this onlology and all imported ontologies for a Property
 * whose ID matches the provided ID. May create a new instance with
 * the appropriate ID in this ontology if none exists.
 * The Property if new is added to the Ontology with the type
 * sets to the passed in type.
 *
 * @param aType The type of the new Property.
 * @param theID The ID of the new Property.
 *
 * @return The matching Property if found or created, otherwise returns void.
 **/
public OGProperty createProperty(OGResource aType,String theID);
/**
 * Creates a Property instance in this ontology with a generated unique ID,
 * sets the Clases type to the passed in type
 * and adds the Property to the ontology.
 *
 * @param aType The type of the new Property.
 *
 * @return The matching Property created, otherwise returns void.
 **/
public OGPropInstance createPropInstance(OGResource aType);
/**
 * Looks for, and possibilly creates, a Property instance in this ontology.
 * Searches this onlology for a Property
 * whose ID matches the provided ID. May create a new instance with
 * the appropriate ID in this ontology if none exists.
 * The Property if new is added to the Ontology with the type
 * sets to the passed in type.
 *
 * @param aType The type of the new Property.
 * @param theID The ID of the new Property.
 *
 * @return The matching Property Instance if found or created, otherwise returns void.
 **/
public OGPropInstance createPropInstance(OGResource aType,String theID);

/**
 * Adds the provided property to the ontology.
 *
 * The provided property should be added so that it will be included in
 * both property and resource collections as appropriate.
 *
 * @param theProperty The property to add to the ontology.
 **/
public void addProperty(OGProperty theProperty);

/**
 * Removes the provided property from the ontology.
 *
 * @param theProperty The property from remove to the ontology.
 **/
public void removeProperty(OGProperty theProperty);

/**
 * Looks for a property in this ontology whose ID matches the provided ID.
 *
 * Differs from lookupProperty in that the imports are not searched.
 *
 * @param theID The ID of the sought property.
 * @return The matching property if found, otherwise returns void.
 **/
public OGProperty getProperty(Object theID);

/**
 * Checks for the existence of  a property in this ontology whose
 * ID matches the provided ID.
 *
 * The property must be in this ontology, not in its imports.
 *
 * @param theID The ID of the sought property.
 * @return true if the property exists in this ontology.
 **/
public boolean containsProperty(Object theID);

/**
 * Provides a iterator for the all of the ontology's properties.
 *
 * @return An iterator over all the ontology's properties
 **/
public Iterator properties();

/**
 * Provides a iterator for only the ontology's properties
 * of the provided type.
 *
 * @param aType The type of properties to include in the iterator.
 * @return An iterator over some of the ontology's properties.
 **/
public Iterator properties(OGResource aType);

/**
 * Provides a Collection of the ontology's properties
 * of the provided type.
 *
 * @param aType The type of properties to include in the iterator.
 * @return A collection of some of the ontology's properties.
 **/
public Collection getProperties(OGResource aType);

/*
 * ---------- Literal Methods --------------
 */

/**
 * Looks for a matching literal in this ontology.
 * Searches this onlology and all imported ontologies for a literal
 * whose ID matches the provided ID.
 *
 * @param theID The ID of the sought literal.
 * @return The matching literal if found, otherwise returns void.
 **/
public OGLiteral lookupLiteral(Object theID);

/**
 * Looks for, and possibilly creates, a literal in this ontology.
 * Searches this onlology and all imported ontologies for a literal
 * whose ID matches the provided ID. May create a new instance with
 * the appropriate ID in this ontology if none exists.
 *
 * @param theID The ID of the sought literal.
 * @param pCreate true causes a new instance to be created if no existing
 * match is found.
 *
 * @return The matching literal if found or created, otherwise returns void.
 **/
public OGLiteral lookupLiteral(Object theID, boolean pCreate);

/**
 * Creates a Literal in this ontology with a generated unique ID
 * and adds the Literal to the ontology.
 *
 * @return The matching Literal created, otherwise returns void.
 **/
public OGLiteral createLiteral();
/**
 * Looks for, and possibilly creates, a Literal in this ontology.
 * Searches this onlology and all imported ontologies for a Literal
 * whose ID matches the provided value. May create a new instance with
 * the appropriate ID in this ontology if none exists.
 * The Literal if new is added to the Ontology.
 *
 * @param theID The ID of the new Literal.
 *
 * @return The matching Literal if found or created, otherwise returns void.
 **/
public OGLiteral createLiteral(String theID);
/**
 * Looks for, and possibilly creates, a Literal in this ontology.
 * Uses the OGLiteral genID method to create the ID for the provided
 * value, then searches this onlology and all imported ontologies for a Literal
 * whose ID matches.  May create a new instance with
 * the appropriate ID in this ontology if none exists and set its value.
 * The Literal if new is added to the Ontology.
 *
 * @param theValue The value of the new Literal.
 *
 * @return The matching Literal if found or created, otherwise returns void.
 **/
public OGLiteral createLiteralByValue(String theValue);
/**
 * Creates a Literal in this ontology with a generated unique ID,
 * sets the Clases type to the passed in type
 * and adds the Literal to the ontology.
 *
 * @param aType The type of the new Literal.
 *
 * @return The matching Literal created, otherwise returns void.
 **/
public OGLiteral createLiteral(OGResource aType);
/**
 * Looks for, and possibilly creates, a Literal in this ontology.
 * Uses the OGLiteral genID method to create the ID for the provided
 * value, then searches this onlology and all imported ontologies for a Literal
 * whose ID matches.  May create a new instance with
 * the appropriate ID and type in this ontology if none exists and sets its value.
 * The Literal if new is added to the Ontology.
 *
 * @param aType The type of the new Literal.
 * @param theValue The value of the new Literal.
 *
 * @return The matching Literal if found or created, otherwise returns void.
 **/
public OGLiteral createLiteralByValue(OGResource aType,String theID);
/**
 * Looks for, and possibilly creates, a Literal in this ontology.
 * Searches this onlology and all imported ontologies for a Literal
 * whose ID matches the provided ID. May create a new instance with
 * the appropriate ID in this ontology if none exists.
 * The Literal if new is added to the Ontology with the type
 * sets to the passed in type.
 *
 * @param aType The type of the new Literal.
 * @param theID The ID of the new Literal.
 *
 * @return The matching Literal if found or created, otherwise returns void.
 **/
public OGLiteral createLiteral(OGResource aType,String theID);

/**
 * Adds the provided literal to the ontology.
 *
 * The provided literal should be added so that it will be included in
 * both  literal and resource collections as appropriate.
 *
 * @param theLiteral The literal to add to the ontology.
 **/
public void addLiteral(OGLiteral theLiteral);

/**
 * Removes the provided literal from the ontology.
 *
 * @param theLiteral The literal to remove to the ontology.
 **/
public void removeLiteral(OGLiteral theLiteral);

/**
 * Looks for a literal in this ontology whose ID matches the provided ID.
 *
 * Differs from lookupLiteral in that the imports are not searched.
 *
 * @param theID The ID of the sought literal.
 * @return The matching literal if found, otherwise returns void.
 **/
public OGLiteral getLiteral(Object theID);

/**
 * Looks for a literal in this ontology whose ID matches the provided ID.
 *
 * The literal must be in this ontology, not in its imports.
 *
 * @param theID The ID of the sought literal.
 * @return true if the literal exists in this ontology.
 **/
public boolean containsLiteral(Object theID);

/**
 * Provides a iterator for the all of the ontology's literals.
 *
 * @return An iterator over all the ontology's literals.
 **/
public Iterator literals();

/**
 * Provides a iterator for only the ontology's literals
 * of the provided type.
 *
 * @param aType The type of literals to include in the iterator.
 * @return An iterator over some of the ontology's literals.
 **/
public Iterator literals(OGResource aType);

/**
 * Provides a Collection of the ontology's literals
 * of the provided type.
 *
 * @param aType The type of literals to include in the iterator.
 * @return A collection of some of the ontology's literals.
 **/
public Collection getLiterals(OGResource aType);

/*
 * ---------- Imports Methods --------------
 */

/**
 * Looks for an imported ontology by ID.
 * Searches this onlology's imports recursively for an imported ontology
 * whose ID matches the provided ID.
 *
 * @param theID The ID of the sought ontology.
 * @return The matching ontology if found, otherwise returns void.
 **/
public OGOntology lookupImport(Object theID);

/**
 * Adds the provided ontology to the ontology's imports.
 * *
 * @param theImport The ontology to add to the imports.
 **/
public void addImport(OGOntology theImport);

/**
 * Removes the provided ontology from the ontology's imports.
 *
 * @param theImport The ontology to remove from the imports.
 **/
public void removeImport(OGOntology theImport);

/**
 * Looks for an imported ontology in this ontology whose ID matches the provided ID.
 *
 * Differs from lookupimport in that the imports are not searched recursively.
 *
 * @param theID The ID of the sought ontology.
 * @return The matching ontology if found, otherwise returns void.
 **/
public OGOntology getImport(Object theID);

/**
 * Accessor for the read only indexed bean property, "imports"
 *
 *
 * @return The imported ontologies of this resources as an array.
 **/
public OGOntology[] getImports();


/**
 * Looks for an imported ontology in this ontology whose ID matches the provided ID.
 *
 * The ontology must be in this ontologies imports. Imports are not searched recursively.
 *
 * @param theID The ID of the sought ontology.
 * @return true if the ontology exists in the imports.
 **/
public boolean containsImport(Object theID);

/**
 * Provides a iterator for the all of the ontology's imports.
 *
 * Only this ontology's imports are included.
 *
 * @return An iterator over all the ontology's imports.
 **/
public Iterator imports();

/*
 * ---------- Resource Methods --------------
 */
 /**
  * Adds the provided resource to the ontology.
  *
  * The provided resource should be added so that it will be included in
  * collections as appropriate.
  *
  * @param theResource The resource to add to the ontology.
  **/
 public void addResource(OGResource theResource);

 /**
  * Removes the provided resource from the ontology.
  *
  * @param theResource The resource removed from the ontology.
  **/
 public void removeResource(OGResource theResource);

/**
 * Looks for a matching resource in this ontology.
 *
 * Searches this onlology and all imported ontologies for a resource
 * whose ID matches the provided ID.
 *
 * Resources are added an removed only throught the
 * specific class, property, liertal mehtods.
 *
 * @param theID The ID of the sought resource.
 * @return The matching resource if found, otherwise returns void.
 **/
public OGResource lookupResource(Object theID);

/**
 * Looks for a resource in this ontology whose ID matches the provided ID.
 *
 * Resources are added an removed only throught the
 * specific class, property, liertal mehtods.
 *
 * @param theID The ID of the sought resource.
 * @return The matching class if found, otherwise returns void.
 **/
public OGResource getResource(Object theID);

/**
 * Looks for a resource in this ontology whose ID matches the provided ID.
 *
 * The resource must be in this ontology, not in its imports.
 *
 * @param theID The ID of the sought resource.
 * @return true if the resource exists in this ontology.
 **/
public boolean containsResource(Object theID);

/**
 * Provides a iterator for the all of the ontology's resources.
 *
 * @return An iterator over all the ontology's resources.
 **/
public Iterator resources();

/**
 * Provides a iterator for only the ontology's resources
 * of the provided type.
 *
 * @param aType The type of resources to include in the iterator.
 * @return An iterator over some of the ontology's resources.
 **/
public Iterator resources(OGResource aType);


/**
 * Provides a Collection of the ontology's resources
 * of the provided type.
 *
 * @param aType The type of resources to include in the iterator.
 * @return A collection of some of the ontology's resources.
 **/
public Collection getResources(OGResource aType);

} //

