/*
 * @(#)OGResource.java
 *
 * OntoAPI
 * Copyright 2003 AT&T Govvernment Solution, Inc.
 * All Rights Reserved.
 *
 */
package com.att.gs.codip.ograph;

import java.util.Collection;
import java.util.Iterator;

import com.att.gs.codip.resource.LogicalResource;

/**
 * Provides shared functionality of Ontology Resources.
 *
 * <p>All instances of resource have a colleciton of Properties.</p>
 *
 * <p>The resource ID is unique and immutable once created. Currently
 * the ID is the resources URI.</p>
 *
 * <p>Each resource exists with in a single XML Name Space.</p>
 *
 *
 **/
public interface OGResource extends LogicalResource {

/*
 * ---------- Basic Access Methods --------------
 */
/**
 * Getter method for the Resources ID.

 * The resource ID is unique and immutable once created. Currently
 * the ID is the resources URI.
 *
 * @returns The ID as a String.
 **/
public String getID ();

/**
 * Setter method for this resource's label string.
 *
 * @param theLabel The label of this resource.
 **/
public void setLabel(String theLabel);
/**
 * Getter method for this resource's label string.
 *
 * @returns  The label of this resource.
 **/
public String getLabel ();

/**
 * Setter method for this resource's comment string.
 *
 * @param theComment The comment for this resource.
 **/
public void setComment(String theComment);
/**
 * Getter method for this resource's comment string.
 *
 * @returns  The comment for this resource.
 **/
public String getComment ();


/*
 * URI/RDF Methods.
 */

/**
 * Getter method for this resource's URI.
 *
 * The URI is immutable and is set when an ontology is
 * built from an RDF model. It is also used when an RDF
 * model is built from the ontology.
 *
 * @returns  The URI string of this resource.
 **/
public String getURI();
/**
 * Getter method for this resource's Namespace.
 *
 * The Namespace is immutable and is set when an ontology is
 * built from an RDF model. It is also used when an RDF
 * model is built from the ontology.
 *
 * @returns  The Namespace string of this resource.
 **/
public String getNamespace();
/**
 * Getter method for this resource's Local Name.
 *
 * The Local Name is immutable and is set when an ontology is
 * built from an RDF model. It is also used when an RDF
 * model is built from the ontology.
 *
 * @returns  The Local Name string of this resource.
 **/
public String getLocalName();

/**
 * Getter method for this resource's Ontology.
 *
 * The Ontology is immutable and is set when the resource is created.
 *
 * @returns  The parent Ontology of this resource.
 **/
public OGOntology getOntology();

/*
 * ---------- Type Methods --------------
 */
//
// ------- Multi-type capabaility.
//
/**
 * Type checker for multi-types.
 *
 * @returns  True if the resource is of type theType.
 **/
public boolean containsType(OGResource theType);

/**
 * Adds the given type to the list of multi-types.
 *
 **/
public void addType(OGResource theType);

/**
 * Removes the given type from the list of multi-types.
 *
 * Use with caution.
 *
 **/
public void removeType(OGResource theType);

/**
 * Provides an iterator over the list of multi-types.
 *
 **/
public Iterator types();

/*
 * ---------- Property Instance Methods --------------
 */

/**
 * Looks for a matching property in this resource.
 * Searches this resource for a property whose ID matches the provided ID.
 *
 * @param theID The ID of the sought property.
 * @return The matching property if found, otherwise returns void.
 **/
public OGPropInstance lookupPropInstance(Object theID);

/**
 * Creates a property instantce and adds it to the resource. THe
 * property is of the provided type, it subject is this resource,
 * and its object is the provided resource.
 * *
 * @param newPropType The type of the property instance.
 * @param theObject The object of the property instance.
 * @return The created property instance.
 **/
public OGPropInstance addNewPropInstance(OGResource newPropType,
											  OGResource theObject);

/**
 * Adds the provided property instance to the resource.
 * *
 * @param theProperty The property to add to the resource.
 **/
public void addPropInstance(OGPropInstance thePropInstance);

/**
 * Removes the provided property instance from the resource.
 *
 * @param thePropInstance The property to remove from the resource.
 **/
public void removePropInstance(OGPropInstance thePropInstance);


/**
 * Looks for a property of this resource whose ID matches the provided ID.
 *
 * Functionally indentical to lookupPropInstance.
 *
 * @param theID The ID of the sought property.
 * @return The matching property if found, otherwise returns void.
 **/
public OGPropInstance getPropInstance(Object theID);

/**
 * Accessor for the read only indexed bean property, "properties"
 *
 *
 * @return The properties of this resources as an array.
 **/
public OGPropInstance[] getPropInstances();


/**
 * Checks for the existence of a property in this resource
 * whose ID matches the provided ID.
 *
 * @param theID The ID of the sought property.
 * @return true if the property exists in this ontology.
 **/
public boolean containsPropInstance(Object theID);


/**
 * Provides a iterator for the all of the resource's properties.
 *
 * @return An iterator over all the resource's properties
 **/
public Iterator propInstances();

/**
 * Looks for a property of this resource that is a particular type.
 *
 * @param aType The type of the sought properties.
 * @return An iterator over some the resource's properties.
 **/
public Iterator propInstances(OGResource aType);

/**
 * Provies a Collection for all of the resource's
 * properties that are a particular type.
 *
 * @param aType The type of the sought properties.
 * @return A collection of some of the resource's properties.
 **/
public Collection getPropInstances(OGResource aType);

} // OGResource
