/*
 * @(#)OntologyManager.java
 *
 * Ontology Graph API,  Copyright 2000, 2001 GRCI All Rights Reserved.
 *
 */
package com.att.gs.codip.ograph;

import java.util.Iterator;
import java.util.Map;


/** **************************
 * This is the interface for the ontology manager class.
 ************************** **/
public interface OntologyManager {

/************************************************
 * Simple Access Methods.
 ************************************************/
/**
 * Getter method for the version information  of this ontology manager.
 *
 * @return The version information  String.
 **/
public String getVersionInfo ();

/**
 * Getter method for the ID of this ontology manager.
 *
 * The ID is immutable after creation.
 *
 * @return The ID String.
 **/
public String getID();

/**
 * Getter method for the default core ontology provided
 * by the ontology manager.
 *
 * It is a static singleton instance.
 *
 *
 * @return The instance of the ontology.
 **/
public OGOntology getDAML();
/**
 * Getter method for the default core OWL ontology provided
 * by the ontology manager.
 *
 * It is a static singleton instance.
 *
 *
 * @return The instance of the ontology.
 **/
public OGOntology getOWL();
/**
 * Getter method for the default core XML Schema Datatype ontology provided
 * by the ontology manager.
 *
 * It is a static singleton instance.
 *
 *
 * @return The instance of the XSD ontology.
 **/
public OGOntology getXSD();
/**
 * Getter method for the XSL Transform file URI. The transform is
 * applied to ontology files by the ontology manager before
 * the ontology is parsed.
 * A null string indicates no transform. The default is no transform.
 *
 *
 * @return The XSL file URI, or null for no transform.
 **/
public String getXSLT();
/**
 * Setter method for the XSL Transform file URI. The transform is
 * applied to ontology files by the ontology manager before
 * the ontology is parsed.
 * A null string indicates no transform. The default is no transform.
 *

 * @parm theXSLT The XSL file URI, or null for no transform.
 **/
public void setXSLT(String theXSLT);

public void setApplyToImports(boolean pTransform);

/********************************************
 * Collection Access Methods.
 *******************************************/
/*
 * ---------- Ontology Methods --------------
 */

/**
 * Looks for an existing instance the specified Ontology by ID.
 *
 * The ID can be either a URI or an alias. First the manged collection
 * is searched by URI. If a matching ontology is found it is returned.
 * Then the colleciotn is search by alias. Again, if a matchin ontology
 * is found it is return.
 *
 * @param theID The ID of the sought ontology.
 * @return The matching ontology if found, otherwise returns void.
 **/
public OGOntology lookupOntology(Object theID);

/**
 * Looks for, and possibilly loads, the specified Ontology by URI.
 * Searches for the onlology whose URI matches the provided URI.
 * Will atempt to load a new instance with the appropriate
 * URI if none exists.
 * Aliases are used however, the URI must be a valid URL
 * for the ontology to be created.
 *
 * @param theID The ID of the sought ontology.
 * @param pCreate true causes a new instance to be created if no existing
 * match is found.
 *
 * @return The matching Ontology if found or created, otherwise returns void.
 **/
public OGOntology lookupOntology(Object theURI, boolean pCreate);

/**
 * Creates a new instance of an ontology with a generated unique URI.
 * The created ontology is empty. It is identified by a unique URI.
 *
 * @return The Ontology if created, otherwise returns null.
 **/
public OGOntology createOntology();
/**
 * Loads or creates an ontology by URI.
 * Searches first for an onlology whose ID matches the provided URI,
 * if no match is found a new instance is created and/or loaded. If
 * a match is found, null is returned.
 * Note that the URI must be a valid URL for the ontology to be loaded.
 *
 * @param theURI The URI of the new ontology.
 *
 * @return The matching ontology if loaded or created, otherwise returns null.
 **/
public OGOntology createOntology(String theURI);
/**
 * If forceReload is true will (re)load the ontology from the URI even if
 * an existing instance is found.
 * Otherwise, behaves identically to createOntology(String theURI);
 *
 * @param theID The ID of the new ontology.
 * @param forceReload If true forces the ontology to be reload from the URI.
 *
 * @return The matching ontology if found or created, otherwise returns void.
 **/
public OGOntology createOntology(String theURI, boolean forceReload);

/**
 * Adds the provided ontology to the managed collection.
 *
 * The provided ontology instance is added only if there is
 * not a currently managed instance with the same URI.
 *
 * @param theOntology The ontology instance to add to the collection.
 **/
public void addOntology(OGOntology theOntology);

/**
 * Morphs the provided ontology & updates the managed collection.
 *
 * When the ontology is morphed any references to the old URI should
 * be updated to the new URI.
 *
 * @param theOntology The ontology to remove from the collection.
 **/
public void morphOntology(OGOntology theOntology,String theURI);


/**
 * Removes the provided ontology from the managed collection.
 *
 * When the ontology is removed from the managed colleciton, all
 * aliases refering to it and all other interanl reference to it
 * are also removed. The removed ontology may be garbage collected
 * if no other references exist.
 *
 * @param theOntology The ontology to remove from the collection.
 **/
public void removeOntology(OGOntology theOntology);

/**
 * Looks for an ontology in the managed colleciton whose URI matches
 * the provided URI.
 *
 * The ontology manager assumes that ontologies are uniquely identified
 * by their URI and therefore maintains a static colletion of ontologies
 * with the URI as key.
 *
 * @param theURI The URI of the sought ontology.
 * @return The matching ontology if found, otherwise returns null.
 **/
public OGOntology getOntology(Object theURI);

/**
 * Checks for the existence of an ontology in the mananged colleciton
 * whose ID matches the provided ID.
 *
 * @param theID The ID of the sought ontology.
 * @return true if the ontology exists in this collection.
 **/
public boolean containsOntology(Object theID);

/**
 * Provides a iterator for the ALL of the managed ontologies.
 *
 * @return An iterator over all the ontologies.
 **/
public Iterator ontologies();

/*
 * ---------- Alias Methods --------------
 */
/**
 * Looks up the URI of an alias in this manager.
 *
 * @param anAlias The alias string.
 * @return The URI of the aliased ontology, or null.
 **/
 public String lookupAlias(String anAlias);

/**
 * Checks for the existence of an alias in this manager.
 *
 * @param anAlias The alias string.
 * @return true if the alias exists.
 **/
 public boolean containsAlias(String anAlias);

/**
 * Adds an alias for the supplied ontology.
 *
 * @param anAlias The alias string.
 * @param theOntology The ontology refered to by the alias string.
 **/
 public void addAlias(String anAlias, OGOntology theOntology);

/**
 * Removes the supplied alias.
 *
 * @param anAlias The alias string.
 **/
 public void removeAlias(String anAlias);

/**
 * Provides an iterator over all aliases known by this manager.
 *
 * @return The iterator.
 **/
 public Iterator aliases();

/**
 * Provides an iterator over all aliases for the spefcified
 * ontology that are known by this manager.
 *
 * @return The iterator.
 **/
 public Iterator aliases(OGOntology theOntology);

/*
 * ---------- Parameter Methods --------------
 */
/**
 * Returns an ontology's parameter map.
 *
 * @param theURI The URI of the ontology.
 * @return a Map object containing the parameters for the ontology.
 **/
public Map getParameters(String theURI);

/**
 * Looks up and returns an ontology's parameter value.
 *
 * @param theURI The URI of the ontology.
 * @param aParameter The name of the parameter.
 * @return The value of the parameter for the ontology.
 **/
public Object lookupParameter(String theURI, String aParameter);

/**
 * Adds or replaces an ontology parameter.
 *
 * @param theURI The URI of the ontology.
 * @param aParameter The name of the parameter to add or update.
 * @param aValue The new value of the parameter.
 **/
public void addParameter(String theURI, String aParameter, Object aValue);

/**
 * Removes an ontology's parameter.
 *
 * @param theURI The URI of the ontology.
 * @param aParameter The name of the parameter to remove.
 **/
public void removeParameter(String theURI, String aParameter);

} // OntoMan
