/*
 * Apollo.java
 *
 * Created on May 5, 2004, 12:37 PM
 */

package cz.cvut.km.io;

import cz.cvut.km.model.*;
import cz.cvut.km.model.iterators.*;
import cz.cvut.km.utils.*;

import java.io.*;
import java.util.*;
import org.w3c.dom.*;

import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;

//Apollo
import edu.ou.kmi.apollo.model.*;
import edu.ou.kmi.apollo.model.types.*;
import edu.ou.kmi.apollo.model.facets.*;
import edu.ou.kmi.apollo.core.*;
import edu.ou.kmi.apollo.plugins.io.xml.*;

/**
 * Apollo I/O using OGraph
 * @author  Petr Aubrecht, Monika Zakova
 */
public class Apollo implements IoInterface {
    
    /** name of property of concepts according to Apollo */
    public static final String APPTYPE = "apollo-type";
    /** property of concepts according to Apollo -- class*/
    public static final String TYPE_CLASS = "class";
    /** property of concepts according to Apollo -- instance*/
    public static final String TYPE_INSTANCE = "instance";
    /** property of concepts according to Apollo -- property (slot)*/
    public static final String TYPE_PROPERTY = "property";
    /** property of concepts according to Apollo -- assignment (value to slot)*/
    public static final String TYPE_ASSIGNMENT = "assignment";
    
    public static final String TYPE_LITERAL = "literal";
    public static final String TYPE_SLOT = "slot";
    
    public static final String SYSTEM = "system:";
    
    // gate-specific concepts
    public Concept cClass = null;
    public Concept cSlot = null;
    public Concept cInstance = null;
    public Concept cLiteral = null;
    
    //TODO: add system namespace as in owl
    
    
    /** Creates a new instance of Apollo. */
    public Apollo() {
    }
    
    /**
     * load a knowledge base from the I/O module.
     * @return the loaded knoledge base
     * @param parameters properties to customise the kb load, must contain <pre>source</pre> for the filename, may contain <pre>namespace</pre> with the namespace of the loaded concepts
     * @throws KMException all other exception are covered by KMException (not existing file, bad format, etc.)
     */
    public cz.cvut.km.model.KnowledgeBase loadKB(java.util.Properties parameters) throws cz.cvut.km.utils.KMException {
        try {
            KnowledgeBase kb = new KnowledgeBase();
            
            //            String aNS = parameters.getProperty("namespace","apollo"); // namespace for the Apollo concepts
            String aNS = parameters.getProperty("namespace","http://www.cvut.cz/apollo#"); // namespace for the Apollo concepts
            String source = parameters.get("source").toString();
            boolean doSystemConcepts = false;
            String doSystemConceptsStr = (String)parameters.get("system-concepts");
            if(doSystemConceptsStr!=null && doSystemConceptsStr.equalsIgnoreCase("true"))
                doSystemConcepts = true;
            if(doSystemConcepts) {
                kb = createSystemOntology();
            }
            
            // init Apollo logging
            edu.ou.kmi.util.Logging.init("Apollo.log");
            // load XML
            ApolloModifiedXMLIOPlugin xmlin = new ApolloModifiedXMLIOPlugin();
            Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(source);
            //silent Apollo while loading
            PrintStream systemOut = System.out;
            System.setOut(new PrintStream(new OutputStream() {
                public void write(int i) {}
            }));
            KOntology onto = xmlin.readOntologyFromXML(doc);//new RepositoryClient());
            System.setOut(systemOut);
            
            kb.setName(onto.getName());
            
            // define all classes
            Iterator classes = onto.getClasses().getIterator();
            while(classes.hasNext()) {
                KClass cls = (KClass)classes.next();
                Concept concept = new Concept(aNS,cls.getName());
                kb.addConcept(concept);
                if(doSystemConcepts) {
                    kb.addRelation(new InstanceOf(concept,cClass));
                }
            }
            
            // define all instances
            Iterator instances = onto.getInstances().getIterator();
            while(instances.hasNext()) {
                KInstance inst = (KInstance)instances.next();
                Concept concept = new Concept(aNS,inst.getFullName());
                kb.addConcept(concept);
                kb.addRelation(new InstanceOf(concept,kb.getConcept(new Concept(aNS,inst.getInstanceOf().getName()))));
                if(doSystemConcepts) {
                    kb.addRelation(new InstanceOf(concept,cInstance));
                }
            }
            
            // add superclasses
            classes = onto.getClasses().getIterator();
            while(classes.hasNext()) {
                KClass cls = (KClass)classes.next();
                Concept concept = kb.getConcept(new Concept(aNS,cls.getName()));
                Iterator superclasses = cls.getDirectSuperClasses().getIterator();
                while(superclasses.hasNext()) {
                    KClass supercls = (KClass)superclasses.next();
                    kb.addRelation(new SubclassOf(concept,kb.getConcept(new Concept(aNS,supercls.getName()))));
                }
            }
            
            // add slot definitions and default values (classes)
            classes = onto.getClasses().getIterator();
            while(classes.hasNext()) {
                KClass cls = (KClass)classes.next();
                String clsName = cls.getName();
                Concept concept = kb.getConcept(new Concept(aNS,clsName));
                Iterator slots = cls.getSlots().getIterator();
                while(slots.hasNext()) {
                    KSlot slot = (KSlot)slots.next();
                    if(slot.isInherited() && !slot.isRedefined())
                        continue;
                    Concept kbSlot = new Concept(aNS,clsName+":"+slot.getName());
                    kb.addConcept(kbSlot);
                    if(doSystemConcepts) {
                        kb.addRelation(new InstanceOf(kbSlot,cSlot));
                    }
                    kb.addRelation(new HasDomain(kbSlot,concept));
                    Iterator slotclasses = slot.getValueClass().getIterator();
                    while(slotclasses.hasNext()) {
                        KClass slotclass = (KClass)slotclasses.next();
                        String rangeClass = slotclass.getName();
                        kb.addRelation(new HasRange(kbSlot,kb.getConcept(new Concept(aNS,rangeClass))));
                    }
                    addDefaultValue(kb,aNS,cls,slot,doSystemConcepts);
                }
            }
            
            // add slot values (instances)
            instances = onto.getInstances().getIterator();
            while(instances.hasNext()) {
                KInstance inst = (KInstance)instances.next();
                Concept concept = new Concept(aNS,inst.getFullName());
                //KObjectMap s = inst.getSlots();
                //if(s.size()>0)
                //    System.out.println(inst.getName());
                Iterator slots = inst.getSlots().getIterator();
                while(slots.hasNext()) {
                    KSlot slot = (KSlot)slots.next();
                    if(!slot.isRedefined())
                        continue;
                    addValues(kb,aNS,inst,slot,doSystemConcepts);
                }
            }
            
            // assign types if using system concepts
            if(doSystemConcepts)
                assignTypes(kb);
            
            return kb;
        } catch(Exception e) {
            throw new cz.cvut.km.utils.KMException("Unable to load Apollo from XML",e);
        }
    }
    
    protected String findSlot(KClass cls, KSlot slot){
        
        KSlot classSlot = cls.findSlot(slot.getName());
        if(classSlot != null){
            if(!classSlot.isInherited() || (classSlot.isInherited() && classSlot.isRedefined())){
                return cls.getFullName()+":"+classSlot.getName();
            }
        }
        
        Iterator it = cls.getDirectSuperClasses().getIterator();
        while(it.hasNext()){
            KClass sup = (KClass)it.next();
            String name = findSlot(sup, slot);
            if(name != null)
                return name;
        }
        
        return null;
    }
    
    /**
     * add values of a slot to the knowledge base
     */
    protected void addValues(KnowledgeBase kb, String aNS, KInstance concept, KSlot slot, boolean doSystemConcepts) {
        String slotName = slot.getName();
        Iterator values = slot.getValue().getIterator();
        
        String fullSlotName = findSlot(concept.getInstanceOf(), slot);
        if(fullSlotName == null){
            //System.out.println(slot.getName()+" not found!");
            return;
        }
        Concept slotConcept = kb.getConcept(new Concept(aNS,fullSlotName));
        
        while(values.hasNext()) {
            KObject value = (KObject)values.next();
            //System.out.println(concept.getFullName() + ":" + slotName + "=" + value.toString());
            Concept source = kb.getConcept(new Concept(aNS,concept.getFullName()));
            Concept target = null;
            Concept assignment = new Concept(aNS,"ASSIGN-"+concept.getFullName()+":"+slotName);
            kb.addConcept(assignment);
            kb.addRelation(new InstanceOf(assignment, slotConcept));
            if(value instanceof KInstance)
                target = kb.getConcept(new Concept(aNS,((KInstance)value).getFullName()));
            else if(value instanceof KClass)
                target = kb.getConcept(new Concept(aNS,((KClass)value).getFullName()));
            else if(value instanceof KString){
                target = new Concept(aNS, ((KString)value).getValue());
                kb.addConcept(target);
                if(doSystemConcepts)
                    kb.addRelation(new InstanceOf(target, cLiteral));
            } else
                System.err.println("!!! unknown target type: "+concept.getFullName()+":"+slotName+"="+value.toString());
            kb.addRelation(new PropertyOf(assignment,source));
            if(target!=null)
                kb.addRelation(new HasValue(assignment,target));
        }
    }
    
    /**
     * add default value of a slot to the knowledge base
     */
    protected void addDefaultValue(KnowledgeBase kb, String aNS, KClass concept, KSlot slot, boolean doSystemConcepts) {
        
        String slotName = slot.getName();
        Iterator values = slot.getValue().getIterator();
        
        Concept slotConcept = kb.getConcept(new Concept(aNS,concept.getFullName()+":"+slotName));
        
        while(values.hasNext()) {
            KObject value = (KObject)values.next();
            //System.out.println(concept.getFullName() + ":" + slotName + "=" + value.toString());
            Concept source = kb.getConcept(new Concept(aNS,concept.getFullName()));
            Concept target = null;
            if(value instanceof KInstance)
                target = kb.getConcept(new Concept(aNS,((KInstance)value).getFullName()));
            else if(value instanceof KClass)
                target = kb.getConcept(new Concept(aNS,((KClass)value).getFullName()));
            else if(value instanceof KString){
                target = new Concept(aNS, ((KString)value).getValue());
                kb.addConcept(target);
                if(doSystemConcepts)
                    kb.addRelation(new InstanceOf(target, cLiteral));
            } else
                System.err.println("!!! unknown target type: "+concept.getFullName()+":"+slotName+"="+value.toString());
            if(target!=null)
                kb.addRelation(new HasValue(slotConcept,target));
        }
    }
    
    
    /**
     * save a knoledge base
     * @param kb the knowledge base to save
     * @param parameters parameters: filename (file name to save to)
     * @throws IOException IO problems
     * @throws KMException all exceptions are covered by KMException
     */
    public void saveKB(cz.cvut.km.model.KnowledgeBase kb, java.util.Properties parameters) throws java.io.IOException, cz.cvut.km.utils.KMException {
        
        String filename = parameters.get("filename").toString(); // filename to store the content
        String prefix = parameters.getProperty("prefix",""); // prefix of names (Apollo doesn't allow classes Class, Slot etc.
        // init Apollo logging
        edu.ou.kmi.util.Logging.init("Apollo.log");

        // for unification purposes -- now classify automatically
        String doSystemConceptsStr = parameters.getProperty("system-concepts","false");
        if(!doSystemConceptsStr.equalsIgnoreCase("true")) {
            classifyKB(kb);
            // save log
            FileWriter fw = new FileWriter(filename+".log");
            fw.write(mapping.getErrorLog());
            fw.close();
        }
        
        // load XML
        String name = kb.getName();
        if(name.length()==0)
            name = "Unnamed"; // Apollo doesn't like ontologies without a name
        KOntology onto = new KOntology(name);//new RepositoryClient());
        
        // save classes
        ConceptIterator it = kb.getConcepts();
        while(it.hasNext()) {
            Concept c = it.nextConcept();
            if(TYPE_CLASS.equals(c.getProperty(APPTYPE)) && !kb.isSystemNamespace(c.getNameSpace())) {
                String cname = prefix+c.getName();
                KClass cls = onto.createClass(cname);
            }
        }
        
        // add superclasses and slots
        it = kb.getConcepts();
        try {
            while(it.hasNext()) {
                Concept c = it.nextConcept();
                //System.out.println("DEBUG: Processing "+c.toDebugString());
                // adding superclasses
                if(TYPE_CLASS.equals(c.getProperty(APPTYPE)) && !kb.isSystemNamespace(c.getNameSpace())) {
                    KClass cls = onto.findClass(prefix+c.getName());
                    Vector relsFrom = kb.getRelationsFrom(c);
                    for(int i = 0; i < relsFrom.size(); i++){
                        Relation rel = (Relation)relsFrom.elementAt(i);
                        if(rel instanceof SubclassOf){
                            Concept target = rel.getTarget();
                            if(kb.isSystemNamespace(target.getNameSpace()))
                                continue;
                            String cname = prefix+target.getName();
                            KClass superClass = onto.findClass(cname);
                            if ( superClass != null ) //TODO: check when occurs
                                cls.addDirectSuperClass(superClass);
                        }
                    }
                    
                    Vector relsTo = kb.getRelationsTo(c);
                    for(int i = 0; i < relsTo.size(); i++){
                        Relation rel = (Relation)relsTo.elementAt(i);
                        // add slots
                        if(rel instanceof HasDomain){
                            Concept prop = rel.getSource();
                            // find slot value type if defined
                            Vector relProp = kb.getRelationsFrom(prop);
                            Concept tcls = null;
                            Concept tval = null;
                            for(int j = 0; j < relProp.size(); j++){
                                Relation r = (Relation)relProp.elementAt(j);
                                if(r instanceof HasRange)
                                    tcls = r.getTarget();
                                else if(r instanceof HasValue)
                                    tval = r.getTarget();
                            }
                            // create slot with value of type instance
                            if(tcls != null && onto.findClass(prefix+tcls.getName()) != null ){
                                KSlot newSlot = cls.createTemplateSlot(getShortName(cls.getName(),prefix+prop.getName()), prefix+tcls.getName());
                                if(tval != null){
                                    KInstance inst = onto.findInstance(tval.getName());
                                    KObjectList list = new KObjectList();
                                    if(inst != null){
                                        list.add(inst);
                                        newSlot.setValue(list);
                                    }
                                }
                                // create slot with type literal and default value
                            } else if(tval != null){
                                KString valueString = new KString(tval.getName());
                                KSlot newSlot = cls.createTemplateSlot(getShortName(cls.getName(),prefix+prop.getName()),
                                valueString.getInstanceOf().getName());
                                KObjectList list = new KObjectList();
                                list.add(valueString);
                                newSlot.setValue(list);
                                
                            } else {
                                KString str = new KString("String");
                                KSlot newSlot = cls.createTemplateSlot(getShortName(cls.getName(),prefix+prop.getName()),
                                str.getInstanceOf().getName());
                                
                                
                            }
                            
                            
                            // add slots with restricted type or default values
                        } else if(rel instanceof PropertyOf){
                            Concept assign = rel.getSource();
                            Concept prop = kb.getInstanceOf(assign);
                            if(prop == null)
                                continue;
                            Vector valRels = kb.getRelationsFrom(assign);
                            for(int j = 0; j < valRels.size(); j++){
                                Relation r = (Relation)valRels.elementAt(j);
                                if(r instanceof HasValue){
                                    Concept val = r.getTarget();
                                    // slot with restricted type
                                    if(onto.findClass(prefix+val.getName()) != null ){
                                        cls.createTemplateSlot(getShortName(cls.getName(),prefix+prop.getName()), prefix+val.getName());
                                        // slot with default values
                                    } else {
                                        Concept type = kb.getInstanceOf(val);
                                        if(type!=null && onto.findClass(prefix+type.getName()) != null ){
                                            KSlot kslot = cls.createTemplateSlot(getShortName(cls.getName(),prefix+prop.getName()), prefix+type.getName());
                                            if(kslot == null)
                                                kslot = cls.findSlot(getShortName(cls.getName(),prefix+prop.getName()));
                                            KObjectList list = kslot.getValue();
                                            if(list == null)
                                                list = new KObjectList();
                                            KInstance kinst = onto.findInstance(getShortName(onto,prefix+val.getName()));
                                            if(kinst == null)
                                                kinst = onto.createInstance(onto.findClass(prefix+type.getName()), getShortName(prefix+type.getName(),prefix+val.getName()));
                                            list.add(kinst);
                                            kslot.setValue(list);
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        } catch(Exception e) {
            throw new KMException("Unable to export ontology to Apollo! "+e,e);
        }
        
        // add instances
        it = kb.getConcepts();
        try {
            while(it.hasNext()) {
                Concept c = it.nextConcept();
                if(TYPE_CLASS.equals(c.getProperty(APPTYPE)) && !kb.isSystemNamespace(c.getNameSpace())){
                    KClass cls = onto.findClass(prefix+c.getName());
                    
                    Vector relsTo = kb.getRelationsTo(c);
                    for(int i = 0; i < relsTo.size(); i++){
                        Relation rel = (Relation)relsTo.elementAt(i);
                        if(rel instanceof InstanceOf){
                            Concept source = rel.getSource();
                            String iname = prefix+source.getName();
                            onto.createInstance(cls, getShortName(cls.getName(),iname));
                            
                        }
                    }
                }
            }
        } catch(Exception e) {
            throw new KMException("Unable to export instances to Apollo! "+e);
        }
        
        
        // add slots with values to instances
        it = kb.getConcepts();
        try {
            while(it.hasNext()) {
                Concept inst = it.nextConcept();
                if(TYPE_INSTANCE.equals(inst.getProperty(APPTYPE)) && !kb.isSystemNamespace(inst.getNameSpace())) {
                    KInstance kinst = onto.findInstance(getShortName(onto,prefix+inst.getName()));
                    if(kinst == null) continue;
                    kinst.updateInstanceSlots();
                    
                    Vector relsTo = kb.getRelationsTo(inst);
                    for(int i = 0; i < relsTo.size(); i++){
                        Relation rel = (Relation)relsTo.elementAt(i);
                        if(rel instanceof PropertyOf){
                            Concept slot = rel.getSource();
                            // skip system defined properties such as type
                            if(kb.isSystemNamespace(slot.getNameSpace()))
                                continue;
                            String sname = prefix+kb.getInstanceOf(slot).getName();
                            KSlot kslot = (KSlot)kinst.getSlots().find(getShortName(onto,sname));
                            
                            // in case slot not defined in the class, add it to the class first
                            if(kslot == null){
                                // TODO: handle properties with range defined only by restriction
                                if(kb.getRangeOf((kb.getInstanceOf(slot))) == null) continue;
                                
                                String className = kb.getRangeOf((kb.getInstanceOf(slot))).getName();
                                KClass cls = onto.findClass(prefix+kb.getInstanceOf(inst).getName());
                                cls.createTemplateSlot(getShortName(cls.getName(),sname), className);
                                kinst.updateInstanceSlots();
                                kslot = (KSlot)kinst.getSlots().find(getShortName(onto,sname));
                            }
                            
                            // add value to the slot
                            Vector relsFrom = kb.getRelationsFrom(slot);
                            for(int j = 0; j < relsFrom.size(); j++){
                                Relation r = (Relation)relsFrom.elementAt(j);
                                if(r instanceof HasValue){
                                    Concept value = r.getTarget();
                                    if(value != null && kslot != null){
                                        KObjectList values = new KObjectList();
                                        // TODO: add support of literals
                                        if(onto.findInstance(getShortName(onto,prefix+value.getName())) == null)
                                            values.add(new KString(value.getName()));
                                        else
                                            values.add(onto.findInstance(getShortName(onto,prefix+value.getName())));
                                        kslot.setValue(values);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        } catch(Exception e){
            e.printStackTrace();
            throw new KMException("Unable to export instance slots to Apollo! "+e);
        }
        
        
        // save ontology to XML
        ByteArrayOutputStream tmpfile = new ByteArrayOutputStream();
        try {
            ApolloModifiedXMLIOPlugin xmlin = new ApolloModifiedXMLIOPlugin();
            xmlin.setStream(tmpfile);
//            PrintStream systemOut = System.out;
//            System.setOut(new PrintStream(new OutputStream() {
//                public void write(int i) {}
//            }));
            xmlin.writeOntology(onto);
//            System.setOut(systemOut);
        } catch(Exception e) {
            e.printStackTrace();
            throw new KMException("Unable to export ontology to XML! "+e);
        }
        
        // save it as XML
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document doc = db.parse(new ByteArrayInputStream(tmpfile.toByteArray()));
            Element root = doc.getDocumentElement();
            
            TransformerFactory tFactory = TransformerFactory.newInstance();
            Transformer transformer = tFactory.newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT,"yes");
            DOMSource source = new DOMSource(root);
            StreamResult result = new StreamResult(new FileWriter(filename));
            transformer.transform(source, result);
        } catch(Exception e) {
            throw new KMException("Unable to parse XML of the ontology! "+e);
        }
        
    }
    
    /** assign Apollo type to the concept
     * @param c concept to be set up
     * @param type name of the apollo type
     * @throws KMException when the concept has already a type and the both differ
     */
    protected void assignType(Concept c, String type) throws KMException {
        String ctype = c.getProperty(APPTYPE);
        if(ctype!=null && !type.equals(type))
            throw new KMException("Unable to classify "+c.toString()+" to "+type+", it is already classified to "+ctype);
        c.setProperty(APPTYPE,type);
        //System.out.println(c.toString()+" -> "+type);
    }
    
    Map superproperties = new HashMap(); // property -> superproperty
    Map propertiesRange = new HashMap(); // property -> Class (range)
    Map propertiesDomain = new HashMap(); // property -> Class (domain)
    /** Solve the subproperty problem
     * @param kb the knowledge base to be solved
     */
    public void solveSubproperty(KnowledgeBase kb) {
        superproperties.clear();
        propertiesRange.clear();
        propertiesDomain.clear();
        // first, collect sub/superproperties
        Vector relationsToRemove = new Vector();
        RelationIterator ri = new RelationIterator(new NonSystemRelationIterator(kb.getRelations()));
        while(ri.hasNext()) {
            Relation rel = ri.nextRelation();
            if(rel instanceof SubclassOf) {
                Concept source = rel.getSource();
                if(source.getProperty(APPTYPE,"").equals(TYPE_PROPERTY)) {
                    Concept target = rel.getTarget();
                    if(target.getProperty(APPTYPE,"").equals(TYPE_PROPERTY)) {
                        // yes, it's subproperty relation
                        superproperties.put(source,target);
                        relationsToRemove.add(rel);
                        // look for domain
                        Vector rels = kb.getRelationsFrom(source);
                        for(int i=0;i<rels.size();i++) {
                            Relation xrel = (Relation)rels.get(i);
                            if(xrel instanceof HasDomain)
                                propertiesDomain.put(source, xrel.getTarget());
                            else if(xrel instanceof HasRange)
                                propertiesRange.put(source, xrel.getTarget());
                        }
                    }
                }
            }
        }
        
        for(int i=0;i<relationsToRemove.size();i++) {
            Relation rel = (Relation)relationsToRemove.get(i);
            Concept subprConcept = rel.getSource();
            Concept range=null, domain=null;
            Concept currProperty;
            // traverse superproperties to find range
            currProperty = subprConcept;
            if((Concept)propertiesRange.get(subprConcept)==null) { // interested inly if not already specified
                while(range==null && currProperty!=null) {
                    range = (Concept)propertiesRange.get(subprConcept);
                    currProperty = (Concept)superproperties.get(currProperty);
                }
            }
            if(!kb.removeRelation(rel))
                ; // System.out.println("!!!!! Unable to remove "+relationsToRemove.get(i));
        }
    }
    
    /**
     * Classify concepts of the knowledge base for Apollo
     * @param kb the knowledge base to classify
     * @throws KMException fails if the classification is not possible
     */
    public void classifyKB(cz.cvut.km.model.KnowledgeBase kb) throws KMException {
        classifyKBByMapping(kb); // classification by Mapping
        solveSubproperty(kb);
    }
    
    /** create set of validation rules
     */
    public Mapping createValidator() {
        Mapping mapping = new Mapping(APPTYPE);
        mapping.addRule(TYPE_CLASS,      SubclassOf.class, TYPE_CLASS);
        mapping.addRule(TYPE_PROPERTY,   HasRange.class,   TYPE_CLASS);
        mapping.addRule(TYPE_PROPERTY,   HasDomain.class,  TYPE_CLASS);
        mapping.addRule(TYPE_ASSIGNMENT, PropertyOf.class, TYPE_INSTANCE);
        mapping.addRule(TYPE_ASSIGNMENT, PropertyOf.class, TYPE_CLASS);
        mapping.addRule(TYPE_ASSIGNMENT, HasValue.class,   TYPE_INSTANCE);
        mapping.addRule(TYPE_ASSIGNMENT, HasValue.class,   TYPE_CLASS);
        mapping.addRule(TYPE_PROPERTY,   HasValue.class,   TYPE_INSTANCE);
        mapping.addRule(TYPE_PROPERTY,   HasValue.class,   TYPE_CLASS);
        mapping.addRule(TYPE_INSTANCE,   InstanceOf.class, TYPE_CLASS);
        mapping.addRule(TYPE_ASSIGNMENT, InstanceOf.class, TYPE_PROPERTY);
        // virtual support of subproperty, is solved by method solveSubproperty
        //mapping.addRule(TYPE_PROPERTY,   SubclassOf.class, TYPE_PROPERTY);
        return mapping;
    }
    
    /**
     * Validate the knowledge base for Apollo
     * @param kb the knowledge base to be checked
     */
    public void validateKB(cz.cvut.km.model.KnowledgeBase kb) {
        Mapping mapping = createValidator();
        mapping.validate(kb);
    }
    
    // Mapping used for classification
    Mapping mapping; 
    /** assign types to concepts
     */
    public void classifyKBByMapping(cz.cvut.km.model.KnowledgeBase kb) throws KMException {
        mapping = createValidator();
        mapping.doMapping(kb);
    }
    
    /** check if the concept is of a particular type
     * @param c concept to check
     * @param type typename (one of TYPE_*)
     */
    protected boolean hasType(Concept c, String type) {
        return type.equals(c.getProperty(APPTYPE));
    }
    
    public KnowledgeBase createSystemOntology() {
        
        KnowledgeBase kb = new KnowledgeBase();
        kb.addSystemNamespace("system:apollo");
        cClass = new Concept("system:apollo","Class");
        kb.addConcept(cClass);
        cSlot = new Concept("system:apollo","Slot");
        kb.addConcept(cSlot);
        cInstance = new Concept("system:apollo","Instance");
        kb.addConcept(cInstance);
        cLiteral = new Concept("system:apollo", "Literal");
        kb.addConcept(cLiteral);
        return kb;
    }
    
    
    protected String getShortName(String className, String name){
        int index = name.indexOf(":");
        if(index > 0){
            String clsName = name.substring(0,index);
            //TODO: there may be problem with prefix - is prefix really needed?
            if(clsName.equals(className))
                name = name.substring(index+1,name.length());
        }
        return name;
    }
    
    protected String getShortName(KOntology onto, String name){
        int index = name.indexOf(":");
        if(index > 0){
            String clsName = name.substring(0,index);
            //TODO: there may be problem with prefix - is prefix really needed?
            if(onto.findClass(clsName) != null)
                name = name.substring(index+1,name.length());
        }
        return name;
    }
    
    
    protected void assignTypes(KnowledgeBase kb) throws KMException {
        ConceptIterator ci = kb.getConcepts();
        Vector slots = new Vector();
        Vector assigns = new Vector();
        while(ci.hasNext()){
            Concept concept = ci.nextConcept();
            Iterator it = kb.getRelationsFrom(concept).iterator();
            while(it.hasNext()){
                Relation rel = (Relation)it.next();
                if(rel instanceof InstanceOf){
                    Concept target = rel.getTarget();
                    if(target.equals(cClass)){
                        assignType(concept, TYPE_CLASS); break;
                    } else if(target.equals(cInstance)){
                        assignType(concept, TYPE_INSTANCE); break;
                    } else if(target.equals(cLiteral)){
                        assignType(concept, TYPE_LITERAL); break;
                    } else if(target.equals(cSlot)){
                        assignType(concept, TYPE_PROPERTY);
                        slots.addElement(concept); break;
                    } else {
                        assigns.addElement(concept);
                    }
                }
            }
        }
        
        // handle assignments, since there is no concept of assignment in system ontology
        for(int i = 0; i < assigns.size(); i++){
            Concept assign = (Concept)assigns.elementAt(i);
            for(int j = 0; j < slots.size(); j++){
                Concept slot = (Concept)slots.elementAt(j);
                if(kb.getRelation(new InstanceOf(assign, slot)) != null){
                    assignType(assign, TYPE_ASSIGNMENT);
                }
            }
        }
    }
    
    public boolean supports(String action) {
        return action.equals(SUPPORTS_LOAD) || action.equals(SUPPORTS_SAVE);
    }
    
    public boolean isApolloType(String str){
        if ( str.equalsIgnoreCase(TYPE_CLASS) || str.equalsIgnoreCase(TYPE_INSTANCE)
        || str.equalsIgnoreCase(TYPE_SLOT) ){
            return true;
        }
        return false;
        
    }
    
    
}
