/*
 * owl.java
 *
 * Created on May 5, 2004, 12:37 PM
 */

package cz.cvut.km.io;

import cz.cvut.km.model.*;
import cz.cvut.km.model.iterators.*;
import cz.cvut.km.utils.*;

import java.io.*;
import java.util.*;
import org.w3c.dom.*;

import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;

/**
 * BabelFish XML I/O using OGraph
 * @author  aubrech
 */
public class BabelFish implements IoInterface {
    
    /** Creates a new instance of a BabelFish I/O. */
    public BabelFish() {
    }
    
    /**
     * load a knowledge base from the I/O module.
     * @return the loaded knoledge base
     * @param parameters properties to customise the kb load, must contain <pre>source</pre> for the filename, may contain <pre>namespace</pre> with the namespace of the loaded concepts
     * @throws KMException all other exception are covered by KMException (not existing file, bad format, etc.)
     */
    public cz.cvut.km.model.KnowledgeBase loadKB(java.util.Properties parameters) throws cz.cvut.km.utils.KMException {
        try {
            String source = parameters.getProperty("filename");
            KnowledgeBase kb = new KnowledgeBase();
            Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(source);
            Element root = doc.getDocumentElement();
            NodeList nss = root.getElementsByTagName("systemns");
            for(int i=0;i<nss.getLength();i++) {
                kb.addSystemNamespace(nss.item(i).getAttributes().getNamedItem("name").getNodeValue());
            }
            NodeList concepts = root.getElementsByTagName("concept");
            for(int i=0;i<concepts.getLength();i++) {
                Element concept = (Element)concepts.item(i);
                loadConceptFromXML(kb, concept);
            }
            NodeList relations = root.getElementsByTagName("relation");
            for(int i=0;i<relations.getLength();i++) {
                Element relation = (Element)relations.item(i);
                loadRelationFromXML(kb, relation);
            }
            
            return kb;
        } catch(Exception e) {
            throw new cz.cvut.km.utils.KMException("Unable to load BabelFish from XML",e);
        }
    }
    
    /** save properties of the GraphEntity
     */
    protected void saveProperties(Element node, GraphEntity ge) {
        Enumeration props = ge.getPropertyKeys();
        while(props.hasMoreElements()) {
            String pname = props.nextElement().toString();
            String value = ge.getProperty(pname);
            Element elem = node.getOwnerDocument().createElement("prop");
            node.appendChild(elem);
            elem.setAttribute("name", pname);
            elem.setAttribute("value", value);
        }
    }
    
    /** save properties of the GraphEntity
     */
    protected void loadProperties(Element node, GraphEntity ge) {
        NodeList props = node.getElementsByTagName("prop");
        for(int i=0;i<props.getLength();i++) {
            Element prop = (Element)props.item(i);
            String name = prop.getAttribute("name");
            String value = prop.getAttribute("value");
            ge.setProperty(name, value);
        }
    }
    
    /** save the Concept to the XML
     */
    protected void saveConceptToXML(Element parent, Concept concept) {
        Element node = parent.getOwnerDocument().createElement("concept");
        parent.appendChild(node);
        node.setAttribute("name", concept.getName());
        node.setAttribute("namespace", concept.getNameSpace());
        node.setAttribute("caption", concept.getCaption());
        saveProperties(node,concept);
    }
    
    /** load the Concept from the XML
     */
    protected void loadConceptFromXML(KnowledgeBase kb, Element node) {
        Concept concept = new Concept();
        concept.setName(node.getAttribute("name"));
        concept.setNameSpace(node.getAttribute("namespace"));
        concept.setCaption(node.getAttribute("caption"));
        loadProperties(node,concept);
        kb.addConcept(concept);
    }
    
    /** save the Relation to the XML
     */
    protected void saveRelationToXML(Element parent, Relation relation) {
        Element node = parent.getOwnerDocument().createElement("relation");
        parent.appendChild(node);
        node.setAttribute("sourcens", relation.getSource().getNameSpace());
        node.setAttribute("sourcename", relation.getSource().getName());
        node.setAttribute("targetns", relation.getTarget().getNameSpace());
        node.setAttribute("targetname", relation.getTarget().getName());
        node.setAttribute("type", relation.getTextArrow());
        saveProperties(node,relation);
    }
    
    /** save the Relation to the XML
     */
    protected void loadRelationFromXML(KnowledgeBase kb, Element node) {
        Relation relation;
        Concept src = kb.getConcept(new Concept(node.getAttribute("sourcens"),node.getAttribute("sourcename")));
        Concept trg = kb.getConcept(new Concept(node.getAttribute("targetns"),node.getAttribute("targetname")));
        String type = node.getAttribute("type");
        if(type.equals(HasDomain.ARROW)) {
            relation = new HasDomain(src,trg);
        } else if(type.equals(HasRange.ARROW)) {
            relation = new HasRange(src,trg);
        } else if(type.equals(HasValue.ARROW)) {
            relation = new HasValue(src,trg);
        } else if(type.equals(InstanceOf.ARROW)) {
            relation = new InstanceOf(src,trg);
        } else if(type.equals(PropertyOf.ARROW)) {
            relation = new PropertyOf(src,trg);
        } else if(type.equals(SubclassOf.ARROW)) {
            relation = new SubclassOf(src,trg);
        } else {
            System.err.println("Error: Unable to load relation "+type+"!!!");
            return;
        }
        loadProperties(node,relation);
        kb.addRelation(relation);
    }
    
    /**
     * save a knoledge base
     * @param kb the knowledge base to save
     * @param parameters parameters: filename (file name to save to)
     * @throws IOException IO problems
     * @throws KMException all exceptions are covered by KMException
     */
    public void saveKB(cz.cvut.km.model.KnowledgeBase kb, java.util.Properties parameters) throws java.io.IOException, cz.cvut.km.utils.KMException {
        //throw new cz.cvut.km.utils.NotImplementedException();
        String filename = parameters.get("filename").toString(); // filename to store the content
        
        try {
            Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
            Element root = doc.createElement("knowledgebase");
            Iterator nss = kb.getSystemNamespaces();
            while(nss.hasNext()) {
                String ns = nss.next().toString();
                Element nse = doc.createElement("systemns");
                nse.setAttribute("name", ns);
                root.appendChild(nse);
            }
            ConceptIterator cIt = kb.getConcepts();
            while(cIt.hasNext()) {
                Concept c = cIt.nextConcept();
                saveConceptToXML(root,c);
            }
            RelationIterator rIt = kb.getRelations();
            while(rIt.hasNext()) {
                Relation r = rIt.nextRelation();
                saveRelationToXML(root,r);
            }
            Transformer trans = TransformerFactory.newInstance().newTransformer();
            trans.setOutputProperty(OutputKeys.INDENT,"yes");
            DOMSource src = new DOMSource(root);
            StreamResult res = new StreamResult(new FileWriter(filename));
            trans.transform(src, res);
        } catch(Exception e) {
            throw new KMException("Unable to export ontology to BabelFish XML! "+e);
        }
    }
    
    public boolean supports(String action) {
        return action.equals(SUPPORTS_LOAD) || action.equals(SUPPORTS_SAVE);
    }
    
}
