/*
 * mapping.java
 *
 * Created on July 27, 2004, 11:41 AM
 */

package cz.cvut.km.utils;

import cz.cvut.km.model.*;
import cz.cvut.km.model.iterators.*;
import java.util.*;

/**
 * Mapping is used for storage of mapping rules in form Class1 --arrow--> Class2.
 * Classes describe concepts in a gate ontology and the rules cover all possible
 * combinations. Mapping then automatically decides, which BabelFish concepts map
 * to which gate ontology concepts.
 * @author  aubrech
 */
public class Mapping {
    protected Vector rules = new Vector();
    // name of the classification property
    protected String propertyName;
    // write what is done?
    protected boolean debugOutput = false;
    
    
    /** Creates a new instance of mapping
     * @param propertyName name of the property saving the classification
     */
    public Mapping(String propertyName) {
        this.propertyName = propertyName;
    }
    
    /** add rule in form CLASS1 -arrow-> CLASS2
     */
    public void addRule(String sourceClass, Class arrow, String targetClass) {
        rules.add(new Rule(sourceClass,arrow,targetClass));
    }
    
    /** validate the knowledge base according to the provided rules
     */
    public void validate(KnowledgeBase kb) {
        RelationIterator ri = new RelationIterator(new NonSystemRelationIterator(kb.getRelations()));
        while(ri.hasNext()) {
            Relation rel = ri.nextRelation();
            Concept source = rel.getSource();
            String stype = source.getProperty(propertyName);
            Concept target = rel.getTarget();
            String ttype = target.getProperty(propertyName);
            if(debugOutput) System.out.print(source.getName()+"("+stype+")\t"+rel.toString()+"("+ttype+")\t"+target.getName()+"\t");
            boolean ok = false;
            for(int i=0;i<rules.size();i++) {
                Rule rule = (Rule)rules.get(i);
                Class cls = rel.getClass();
                if(cls==rule.arrow) {
                    if(rule.sourceClass.equals(stype)&&rule.targetClass.equals(ttype))
                        ok = true;
                }
            }
            if(ok) {
                if(debugOutput) System.out.println("OK");
            } else {
                if(debugOutput) System.out.println("ERROR!!! ***************************");
            }
        }
    }
    
    Vector ruleItems = null; // vector of MapItems
    //Map no
    
    /** assign classes to the concepts according to the rules
     */
    public int doMapping(KnowledgeBase kb) {
        // first prepare rule items
        fillRuleItems(kb);
        // sort the content, more restrictive rules are placed first
        Collections.sort(ruleItems, new RulesComparator());
        // the iterate the items to safisfy needs
        return search();
    }
    
    /** Comparator used for sorting rules in order to speed-up the mapping.
     *  Sorts rules according to number of possibilities.
     */
    protected class RulesComparator implements java.util.Comparator {
        public int compare(Object o1, Object o2) {
            MapItem mi1 = (MapItem)o1;
            MapItem mi2 = (MapItem)o2;
            if(mi2.numChoices!=mi1.numChoices)
                return mi1.numChoices - mi2.numChoices;
            return mi1.from.getName().compareTo(mi2.to.getName());
        }
    }
    
    
    protected class MapItem {
        public Relation relation; // for debug purposes only, the originating relation
        public Concept from, to; // indices to the array of nodes
        public boolean assignedFrom=false, assignedTo=false; // from/to was assigned by this rule?
        public Vector fromChoices=new Vector(), toChoices=new Vector(); // class names for nodes, option is (fromChoices[i],toChoices[i])
        public int currChoice=-1; // current choice
        public int numChoices; // number of available choices (shortcut)
        /** return string identification of the current choice
         */
        public String debugCurrChoice() {
            return fromChoices.get(currChoice)+" -> "+toChoices.get(currChoice)+": "+relation.toDebugString();
        }
        /** try the next choice
         */
        public boolean tryNext() {
            // first cleanup
            if(currChoice>=0 && (assignedFrom || assignedTo) ) {
                //if(debugOutput) System.out.println("Rejecting "+debugCurrChoice());
            }
            if(assignedFrom) {
                assignedFrom = false;
                //                System.out.println("Unbound "+from.toString()+" from "+from.getProperty(propertyName));
                from.remove(propertyName);
            }
            if(assignedTo) {
                assignedTo = false;
                //                System.out.println("Unbound "+to.toString()+" from "+to.getProperty(propertyName));
                to.remove(propertyName);
            }
            
            boolean ok = false; // both sides are valid
            String newFromProperty = null;
            String newToProperty = null;
            while(!ok && currChoice<fromChoices.size()-1) {
                ok = true; // more probable, set to false if there is mismatch
                // move the pointer
                currChoice++;
                // well, there is -- go further
                //System.out.println("Trying "+from.toString()+"("+fromChoices.get(currChoice)+") -> "+to.toString()+"("+toChoices.get(currChoice)+")");
                //if(debugOutput) System.out.println("Trying "+debugCurrChoice());
                
                // now check the concepts whether they have the right classes
                // check from
                String currFrom = fromChoices.get(currChoice).toString(); // have to be
                String fromProperty = from.getProperty(propertyName); // is
                newFromProperty = currFrom; // will be
                if(fromProperty==null) {
                    assignedFrom = true;
                } else if(!currFrom.equals(fromProperty))
                    ok = false;
                
                // check to
                String currTo = toChoices.get(currChoice).toString(); // have to be
                String toProperty = to.getProperty(propertyName); // is
                newToProperty = currTo; // will be
                if(toProperty==null) {
                    assignedTo = true;
                } else if(!currTo.equals(toProperty))
                    ok = false;
            }
            if(ok) {
                if(assignedFrom) {
                    from.setProperty(propertyName, newFromProperty);
                    //System.out.println("Bound "+from.toString()+" to "+fromChoices.get(currChoice));
                }
                if(assignedTo) {
                    to.setProperty(propertyName, newToProperty);
                    //System.out.println("Bound "+to.toString()+" to "+toChoices.get(currChoice));
                }
                return true;
            } else {
                currChoice = -1; // reset
                //if(debugOutput) System.out.println("Unable to classify whole "+relation.toDebugString());
                return false;
            }
        }
    }
    
    Stack errorPoints = new Stack();
    /** search the best coverage
     */
    public int search() {
        errorPoints.clear();
        int maxLevel = -1;
        int n = ruleItems.size();
        int maxi;
        int i=0;
        do {
            i=0; maxi = 0;
            maxLevel++;
            if(debugOutput) System.out.println("Level = "+maxLevel);
            int level = 0;
            while(i>=0 && i<n) {
                if(i>maxi) maxi = i;
                //if(debugOutput) System.out.println("== max "+maxi+", curr "+i+" == "+level+"/"+maxLevel);
                MapItem mi = (MapItem)ruleItems.get(i);
                boolean ok = mi.tryNext();
                if(ok) {
                    i++; // ok, go further
                } else {
                    if(level<maxLevel) { // reached max number of errors?
                        level++; // no, increase level and continue
                        errorPoints.push(new Integer(i));
                        i++;
                    } else {
                        // remove error point if this on top
                        if(errorPoints.size()>0 && ((Integer)errorPoints.peek()).intValue()==i) {
                            errorPoints.pop();
                            level--;
                        }
                        i--; // yes, go back
                    }
                }
            }
        } while(i<0);
        
        if(debugOutput && maxLevel>0) {
            System.out.println("******   Mapping FINISHED OK - number of error rules: "+maxLevel);
            for(int y=0;y<errorPoints.size();y++)
                System.out.println("Ignored: "+getRule(y));
        }
        
        return maxLevel;
    }
    
    /** get String representation of rule at position y
     */
    public String getRule(int y) {
        return ((MapItem)ruleItems.get(((Integer)errorPoints.get(y)).intValue())).relation.toDebugString();
    }
    
    public String getErrorLog() {
        StringBuffer sb = new StringBuffer();
        for(int y=0;y<errorPoints.size();y++) {
            sb.append("Ignored: ");
            sb.append(getRule(y));
            sb.append("\n");
        }
        return sb.toString();
    }
    
    Map ruleClassesCount = new HashMap();
    /** fill rule items to ruleItems
     */
    public void fillRuleItems(KnowledgeBase kb) {
        // first count
        for(int i=0;i<rules.size();i++) {
            Rule rule = (Rule)rules.get(i);
            Class cls = rule.arrow;//.getClass();
            Integer num = (Integer)ruleClassesCount.get(cls);
            if(num==null)
                ruleClassesCount.put(cls,new Integer(1));
            else
                ruleClassesCount.put(cls,new Integer(num.intValue()+1));
        }
        
        ruleItems = new Vector();
        
        RelationIterator ri = new RelationIterator(new NonSystemRelationIterator(kb.getRelations()));
        while(ri.hasNext()) {
            Relation rel = ri.nextRelation();
            Class cls = rel.getClass();
            MapItem mi = new MapItem();
            mi.from = rel.getSource();
            mi.to = rel.getTarget();
            mi.relation = rel;
            Integer numChoicesI = ((Integer)ruleClassesCount.get(cls));
            if(numChoicesI==null)
                mi.numChoices = 0;
            else
                mi.numChoices = numChoicesI.intValue();
            ruleItems.add(mi);
            for(int i=0;i<rules.size();i++) {
                Rule rule = (Rule)rules.get(i);
                if(cls==rule.arrow) {
                    mi.fromChoices.add(rule.sourceClass);
                    mi.toChoices.add(rule.targetClass);
                }
            }
        }
    }
    
    /** simple internal structure storing the rule
     */
    public class Rule {
        public String sourceClass;
        public Class arrow;
        public String targetClass;
        public Rule(String sourceClass, Class arrow, String targetClass) {
            this.sourceClass = sourceClass;
            this.arrow = arrow;
            this.targetClass = targetClass;
        }
    }
}
